import re, sys

class Triedicka:
    def __init__(self, verbose = False):
        self.orig_instrukcie = []
        self.instrukcie = []
        self.vstupy = []
        self.verbose = verbose

    def spracuj_vstup(self, hodnoty):
        self.vstupy = hodnoty
        self.n = len( self.vstupy )
        assert self.n > 0, 'Subor so vstupmi neobsahuje ziadne vstupy.'
        self.v = len( self.vstupy[0] )
        assert all( len(x) == self.v for x in self.vstupy ), 'Niektore jadra maju viac vstupov ako ine.'

        if self.verbose:
            print(f'NACITANY VSTUP (jadier: {self.n}, vstupov pre kazde jadro: {self.v})')
            print()

    def nacitaj_vstup(self, subor_so_vstupmi):
        hodnoty = []
        try:
            for line in open(subor_so_vstupmi).readlines():
                riadok = [ int(x) for x in line.split() ]
                assert all( -(2**63) <= x < 2**63 for x in riadok )
                hodnoty.append(riadok)
        except:
            assert False, 'Subor so vstupmi nie je syntakticky korektny.'
        self.spracuj_vstup(hodnoty)

    def nacitaj_instrukcie(self, subor_s_programom):
        assert self.n > 0, 'Kvoli kontrole instrukcii je nutne nacitat najskor vstup a az potom program.'
        self.orig_instrukcie = []
        self.instrukcie = []
        mena_instrukcii = {}
        klasicke_instrukcie = { 'sort':1, 'id':0, 'copy':1, 'left':1, 'right':1, '+':2, '-':2, '*':2, '/':2,
            '%':2, '|':2, '&':2, '^':2, '!': 1, '=':2, '!=':2, '<':2, '>':2, '<=':2, '>=':2, 'if':3 }

        for line in open(subor_s_programom).readlines():
            # zmaz komentar
            if '#' in line:
                line = line[ : line.index('#') ]

            # rozbi na tokeny a preskoc prazdne riadky
            tokeny = line.split()
            if len(tokeny) == 0: continue

            # skontroluj, ci je prvy token label
            label = None
            if tokeny[0][-1] == ':':
                label = tokeny[0][:-1]
                assert label != '', 'Nazov instrukcie nesmie byt prazdny.'
                assert not re.match(r'^[0-9]*$', label), 'Nazov instrukcie nesmie pozostavat iba z cifier.'
                tokeny = tokeny[1:]
            if len(tokeny) == 0: assert False, f'Za menom {label} nenasleduje ziadna instrukcia.'

            # spracujeme samotnu instrukciu
            if tokeny[0] == 'const':
                assert len(tokeny) == 2, f'Instrukcia {tokeny}: nespravny pocet argumentov.'
                try:
                    x = int( tokeny[1] )
                    assert -(2**63) <= x < 2**63
                except:
                    assert False, f'Instrukcia {tokeny}: argument nie je konstanta z povoleneho rozsahu.'
                tokeny[1] = x

            elif tokeny[0] == 'input':
                assert len(tokeny) == 2, f'Instrukcia {tokeny}: nespravny pocet argumentov.'
                try:
                    x = int( tokeny[1] )
                    assert 1 <= x <= self.v
                except:
                    assert False, f'Instrukcia {tokeny}: argument nie je indexom existujuceho vstupu.'
                tokeny[1] = x
            
            else:
                assert tokeny[0] in klasicke_instrukcie, f'Instrukcia {tokeny}: neznamy typ instrukcie.'
                argc = klasicke_instrukcie[ tokeny[0] ]
                assert len(tokeny) == 1+argc, f'Instrukcia {tokeny}: nespravny pocet argumentov.'
                for i in range(1, len(tokeny)):
                    try:
                        x = int( tokeny[i] )
                    except:
                        x = None
                    
                    if x is None:
                        assert tokeny[i] in mena_instrukcii, f'Instrukcia {tokeny}: neznamy nazov skorsej instrukcie.'
                        x = mena_instrukcii[ tokeny[i] ]
                    
                    assert 1 <= x <= len(self.instrukcie), f'Instrukcia {tokeny}: neplatne cislo skorsej instrukcie.'
                    tokeny[i] = x

            self.orig_instrukcie.append( line.strip() )
            self.instrukcie.append( tokeny )
            if self.verbose:
                print(f'Pridana instrukcia {len(self.instrukcie)}: {tokeny}.')
            if label is not None:
                mena_instrukcii[label] = len(self.instrukcie)
                if self.verbose:
                    print(f'Meno "{label}" odteraz ukazuje na instrukciu {mena_instrukcii[label]}.')

        if self.verbose:
            print(f'NACITANY PROGRAM (instrukcii: {len(self.instrukcie)})')
            print()

    def operacia(self, operator, x, y):
        # delenie nulou ako specialny pripad
        if operator in ['/', '%'] and y == 0:
            return 0
        # upravime operator na pythonovy a vypocitame operaciu
        if operator == '/': operator = '//'
        if operator == '=': operator = '=='
        z = eval(f'int( ({x}) {operator} ({y}) )')
        # upravime na spravnu zvyskovu triedu mod 2^64 ak by v int64 nastalo pretecenie
        z %= 2**64
        if z >= 2**63: z -= 2**64
        return z

    def simuluj(self):
        jadra = list( range( self.n ) )
        pamate_jadier = []

        for instr_id in range(len(self.instrukcie)):
            orig_instrukcia = self.orig_instrukcie[instr_id]
            instrukcia = self.instrukcie[instr_id]

            if self.verbose:
                print()
                print(f'Vykonavam instrukciu {instr_id+1}: "{orig_instrukcia}"')
                print(f'Interna reprezentacia instrukcie: {instrukcia}')
            if instrukcia[0] == 'const':
                pamate_jadier.append( [ instrukcia[1] for _ in range( self.n ) ] )
            elif instrukcia[0] == 'input':
                pamate_jadier.append( [ self.vstupy[i][instrukcia[1]-1] for i in range( self.n ) ] )
            elif instrukcia[0] == 'id':
                pamate_jadier.append( list( range( self.n ) ) )
            elif instrukcia[0] == 'copy':
                pamate_jadier.append( pamate_jadier[instrukcia[1]-1] )
            elif instrukcia[0] == 'left':
                vystupy = [ None for _ in range( self.n ) ]
                for i in range( self.n ):
                    ja, sused = jadra[i], jadra[ (i-1) % self.n ]
                    vystupy[ja] = pamate_jadier[instrukcia[1]-1][sused]
                pamate_jadier.append(vystupy)
            elif instrukcia[0] == 'right':
                vystupy = [ None for _ in range( self.n ) ]
                for i in range( self.n ):
                    ja, sused = jadra[i], jadra[ (i+1) % self.n ]
                    vystupy[ja] = pamate_jadier[instrukcia[1]-1][sused]
                pamate_jadier.append(vystupy)
            elif instrukcia[0] == '!':
                vystupy = [ None for _ in range( self.n ) ]
                for i in range( self.n ):
                    x = pamate_jadier[ instrukcia[1]-1 ][i]
                    vystupy[i] = 0 if x else 1
                pamate_jadier.append(vystupy)
            elif instrukcia[0] in ['+', '-', '*', '/', '%', '|', '&', '^', '=', '!=', '<', '>', '<=', '>=']:
                vystupy = [ None for _ in range( self.n ) ]
                for i in range( self.n ):
                    x = pamate_jadier[ instrukcia[1]-1 ][i]
                    y = pamate_jadier[ instrukcia[2]-1 ][i]
                    vystupy[i] = self.operacia( instrukcia[0], x, y )
                pamate_jadier.append(vystupy)
            elif instrukcia[0] == 'if':
                vystupy = [ None for _ in range( self.n ) ]
                for i in range( self.n ):
                    x = pamate_jadier[ instrukcia[1]-1 ][i]
                    y = pamate_jadier[ instrukcia[2]-1 ][i]
                    z = pamate_jadier[ instrukcia[3]-1 ][i]
                    vystupy[i] = y if x != 0 else z
                pamate_jadier.append(vystupy)
            elif instrukcia[0] == 'sort':
                vystupy = [ None for _ in range( self.n ) ]
                pole = []
                x = instrukcia[1]-1
                for i in range( self.n ):
                    j = jadra[i]
                    pole.append( ( pamate_jadier[x][j], i, j ) )
                    vystupy[j] = pamate_jadier[x][j]
                pole.sort()
                jadra = [ _[2] for _ in pole ]
                pamate_jadier.append(vystupy)
            else:
                assert False, f'Neznama instrukcia {instrukcia} (toto nemalo nastat).'

            if self.verbose:
                results = [ f'{x}:{pamate_jadier[-1][x]}' for x in jadra ]
                print(f'Vysledok instrukcie: {results}')

        return (jadra, pamate_jadier)

if __name__ == '__main__':
    try:
        argumenty = sys.argv[1:]
        verbose = any( x in argumenty for x in [ '-v', '--verbose' ] )
        subory = [ x for x in argumenty if x not in [ '-v', '--verbose' ] ]

        interpreter = Triedicka( verbose )
        assert len(subory) == 2, 'Ako argumenty interpretera ocakavam dva subory: vstup a program.'
        interpreter.nacitaj_vstup(subory[0])
        interpreter.nacitaj_instrukcie(subory[1])
        jadra, pamate_jadier = interpreter.simuluj()
        if verbose: print()
        print(f'VYSLEDNE PORADIE JADIER (zlava doprava): {jadra}')
        results = [ pamate_jadier[-1][x] for x in jadra ]
        print(f'VYSTUPY JADIER (zlava doprava): {results}')
        print(f'VYSTUPY JADIER (podla ID): {pamate_jadier[-1]}')
    except Exception as error_message:
        print(error_message)

